@@ -1,11 +1,10 @@
 import os
 from distutils import log
-from distutils.command.check import check
 from distutils.command.clean import clean
 
 from setuptools.command.build_ext import build_ext
 from setuptools.command.install import install
-from distutils.command.sdist import sdist
+from setuptools.command.sdist import sdist
 import sys
 import subprocess
 
@@ -20,15 +19,18 @@ def add_rust_extension(dist):
     sdist_options = sdist_base_class.user_options.copy()
     sdist_boolean_options = sdist_base_class.boolean_options.copy()
     sdist_negative_opt = sdist_base_class.negative_opt.copy()
-    sdist_options.extend([
-        ('vendor-crates', None,
-         "vendor Rust crates"),
-        ('no-vendor-crates', None,
-         "don't vendor Rust crates."
-         "[default; enable with --vendor-crates]"),
-    ])
-    sdist_boolean_options.append('vendor-crates')
-    sdist_negative_opt['no-vendor-crates'] = 'vendor-crates'
+    sdist_options.extend(
+        [
+            ("vendor-crates", None, "vendor Rust crates"),
+            (
+                "no-vendor-crates",
+                None,
+                "don't vendor Rust crates." "[default; enable with --vendor-crates]",
+            ),
+        ]
+    )
+    sdist_boolean_options.append("vendor-crates")
+    sdist_negative_opt["no-vendor-crates"] = "vendor-crates"
 
     class sdist_rust_extension(sdist_base_class):
         user_options = sdist_options
@@ -39,8 +41,7 @@ def add_rust_extension(dist):
             super().initialize_options()
             self.vendor_crates = 0
 
-        def get_file_list(self):
-            super().get_file_list()
+        def make_distribution(self):
             if self.vendor_crates:
                 manifest_paths = []
                 for ext in self.distribution.rust_extensions:
@@ -51,9 +52,7 @@ def add_rust_extension(dist):
                     self.mkpath(dot_cargo_path)
                     cargo_config_path = os.path.join(dot_cargo_path, "config.toml")
                     vendor_path = os.path.join(dot_cargo_path, "vendor")
-                    command = [
-                        "cargo", "vendor"
-                    ]
+                    command = ["cargo", "vendor"]
                     # additional Cargo.toml for extension 1..n
                     for extra_path in manifest_paths[1:]:
                         command.append("--sync")
@@ -65,64 +64,79 @@ def add_rust_extension(dist):
                     # See https://docs.rs/clap/latest/clap/struct.Arg.html#method.multiple for detail
                     command.extend(["--manifest-path", manifest_paths[0], vendor_path])
                     cargo_config = subprocess.check_output(command)
-                    base_dir_bytes = base_dir.encode(sys.getfilesystemencoding()) + os.sep.encode()
-                    if os.sep == '\\':
+                    base_dir_bytes = (
+                        base_dir.encode(sys.getfilesystemencoding()) + os.sep.encode()
+                    )
+                    if os.sep == "\\":
                         # TOML escapes backslash \
                         base_dir_bytes += os.sep.encode()
-                    cargo_config = cargo_config.replace(base_dir_bytes, b'')
+                    cargo_config = cargo_config.replace(base_dir_bytes, b"")
                     if os.altsep:
-                        cargo_config = cargo_config.replace(base_dir_bytes + os.altsep.encode(), b'')
+                        cargo_config = cargo_config.replace(
+                            base_dir_bytes + os.altsep.encode(), b""
+                        )
 
                     # Check whether `.cargo/config`/`.cargo/config.toml` already exists
                     existing_cargo_config = None
-                    for filename in (f".cargo{os.sep}config", f".cargo{os.sep}config.toml"):
-                        if filename in self.filelist.allfiles:
+                    for filename in (
+                        f".cargo{os.sep}config",
+                        f".cargo{os.sep}config.toml",
+                    ):
+                        if filename in self.filelist.files:
                             existing_cargo_config = filename
                             break
                     if existing_cargo_config:
-                        cargo_config_path = os.path.join(base_dir, existing_cargo_config)
+                        cargo_config_path = os.path.join(
+                            base_dir, existing_cargo_config
+                        )
                         # Append vendor config to original cargo config
                         with open(existing_cargo_config, "rb") as f:
-                            cargo_config = f.read() + b'\n' + cargo_config
+                            cargo_config = f.read() + b"\n" + cargo_config
 
                     with open(cargo_config_path, "wb") as f:
                         f.write(cargo_config)
                     self.filelist.append(vendor_path)
                     self.filelist.append(cargo_config_path)
+            super().make_distribution()
+
     dist.cmdclass["sdist"] = sdist_rust_extension
 
-    build_ext_base_class = dist.cmdclass.get('build_ext', build_ext)
+    build_ext_base_class = dist.cmdclass.get("build_ext", build_ext)
+    build_ext_options = build_ext_base_class.user_options.copy()
+    build_ext_options.append(("target", None, "Build for the target triple"))
 
     class build_ext_rust_extension(build_ext_base_class):
+        user_options = build_ext_options
+
+        def initialize_options(self):
+            super().initialize_options()
+            self.target = os.getenv("CARGO_BUILD_TARGET")
+
         def run(self):
             if self.distribution.rust_extensions:
                 log.info("running build_rust")
                 build_rust = self.get_finalized_command("build_rust")
                 build_rust.inplace = self.inplace
                 build_rust.plat_name = self.plat_name
+                build_rust.target = self.target
+                build_rust.verbose = self.verbose
                 build_rust.run()
 
             build_ext_base_class.run(self)
-    dist.cmdclass['build_ext'] = build_ext_rust_extension
 
-    clean_base_class = dist.cmdclass.get('clean', clean)
+    dist.cmdclass["build_ext"] = build_ext_rust_extension
+
+    clean_base_class = dist.cmdclass.get("clean", clean)
 
     class clean_rust_extension(clean_base_class):
         def run(self):
             clean_base_class.run(self)
             if not self.dry_run:
                 self.run_command("clean_rust")
-    dist.cmdclass['clean'] = clean_rust_extension
 
-    check_base_class = dist.cmdclass.get('check', check)
+    dist.cmdclass["clean"] = clean_rust_extension
 
-    class check_rust_extension(check_base_class):
-        def run(self):
-            check_base_class.run(self)
-            self.run_command("check_rust")
-    dist.cmdclass["check"] = check_rust_extension
-
-    install_base_class = dist.cmdclass.get('install', install)
+    install_base_class = dist.cmdclass.get("install", install)
 
     # this is required because, install directly access distribution's
     # ext_modules attr to check if dist has ext modules
@@ -145,7 +159,9 @@ def add_rust_extension(dist):
                     if not self.distribution.entry_points:
                         self.distribution.entry_points = {"console_scripts": scripts}
                     else:
-                        ep_scripts = self.distribution.entry_points.get("console_scripts")
+                        ep_scripts = self.distribution.entry_points.get(
+                            "console_scripts"
+                        )
                         if ep_scripts:
                             for script in scripts:
                                 if script not in ep_scripts:
@@ -161,13 +177,22 @@ def add_rust_extension(dist):
 
             # restore ext_modules
             self.distribution.ext_modules = ext_modules
+
     dist.cmdclass["install"] = install_rust_extension
 
     if bdist_wheel is not None:
         bdist_wheel_base_class = dist.cmdclass.get("bdist_wheel", bdist_wheel)
+        bdist_wheel_options = bdist_wheel_base_class.user_options.copy()
+        bdist_wheel_options.append(("target", None, "Build for the target triple"))
 
         # this is for console entries
         class bdist_wheel_rust_extension(bdist_wheel_base_class):
+            user_options = bdist_wheel_options
+
+            def initialize_options(self):
+                super().initialize_options()
+                self.target = os.getenv("CARGO_BUILD_TARGET")
+
             def finalize_options(self):
                 scripts = []
                 for ext in self.distribution.rust_extensions:
@@ -177,7 +202,9 @@ def add_rust_extension(dist):
                     if not self.distribution.entry_points:
                         self.distribution.entry_points = {"console_scripts": scripts}
                     else:
-                        ep_scripts = self.distribution.entry_points.get("console_scripts")
+                        ep_scripts = self.distribution.entry_points.get(
+                            "console_scripts"
+                        )
                         if ep_scripts:
                             for script in scripts:
                                 if script not in ep_scripts:
@@ -201,15 +228,34 @@ def add_rust_extension(dist):
                     macos_target = os.getenv("MACOSX_DEPLOYMENT_TARGET")
                     if macos_target is None:
                         # Example: macosx_11_0_arm64
-                        macos_target = '.'.join(plat.split("_")[1:3])
+                        macos_target = ".".join(plat.split("_")[1:3])
                     plat = calculate_macosx_platform_tag(
-                        self.bdist_dir,
-                        "macosx-{}-universal2".format(macos_target)
+                        self.bdist_dir, "macosx-{}-universal2".format(macos_target)
                     )
                 return python, abi, plat
+
         dist.cmdclass["bdist_wheel"] = bdist_wheel_rust_extension
 
 
+def patch_distutils_build():
+    """Patch distutils to use `has_ext_modules()`
+
+    See https://github.com/pypa/distutils/pull/43
+    """
+    from distutils.command import build as _build
+
+    class build(_build.build):
+        def finalize_options(self):
+            build_lib_user_specified = self.build_lib is not None
+            super().finalize_options()
+            if not build_lib_user_specified:
+                if self.distribution.has_ext_modules():
+                    self.build_lib = self.build_platlib
+                else:
+                    self.build_lib = self.build_purelib
+
+    _build.build = build
+
 
 def rust_extensions(dist, attr, value):
     assert attr == "rust_extensions"
@@ -220,4 +266,5 @@ def rust_extensions(dist, attr, value):
     )
 
     if dist.rust_extensions:
+        patch_distutils_build()
         add_rust_extension(dist)
